\ asm.f -- ANS Forth assembler (postfix)        28 Mar 2000  tcn
\
\ Designed to compile the Forth-OS from any ANS Forth compiler.
\
\ by Thomas Novelli, 1999-2000
\ Based on the assembler from Pygmy Forth, by Frank Sargent.
\
\ Special notations:
\   ta - Target address

\ TBD:
\   In based/indexed addressing, generate 8-bit displacements
\   (mod=01) or no displacement (mod=00) if DISP=0. See modDISP,

\ "Extraneous" things to do: (only if needed)
\   INCBIN
\   MOV CRx
\   2,3-operand IMUL
\   Do more error checking? (wrong number/type of operands, etc)
\   Support for running under 16-bit Forths (use double cells)

\ Output options:
\ 1. Assemble to blocks (or blockfiles)
\ 2. Assemble to ALLOCATEd memory (and save to file)





ONLY FORTH DEFINITIONS
VOCABULARY ASSEMBLER   ALSO ASSEMBLER DEFINITIONS
HEX    : OCTAL 8 BASE ! ;

variable MODE  ( 0=16-bit 1=32-bit)
: USE16  0 MODE ! ;   : USE32  1 MODE ! ;
: tW ( - n) 2 MODE @ LSHIFT ;  ( Target CELL size, in chars)

variable ORIGIN \ start of code (when loaded on target)
variable -ORG   \ difference
variable ASM0   \ start of code
variable >ASM   \ assembly pointer -- w/rt ASM0 (starts at 0)
variable OUTPUT \ 0=memory, 1=blocks



: $' ( - a  Current address w/rt ASM0)  >ASM @ ;
: $  ( - a  Current address w/rt ORIGIN)  >ASM @  -ORG @ + ;
: ORG ( a -)  DUP ORIGIN !  $' - -ORG ! ;
: T>H ( ta - a)  -ORG @ - ;  \ Note: Invalidated by ORG

( Assemble to memory -- ORG = start address, by default)
: ASM-TO ( a -) DUP ORG  ASM0 !  0 >ASM !  0 OUTPUT ! ;
: MEM-db! ( n ta -) ASM0 @ + C! ;

HEX
( Assemble to blocks -- ORG = 0, by default)
: ASM-BLOCK ( start_block -)
  0a LSHIFT ASM0 !  0 ORG  0 >ASM !  1 OUTPUT ! ;
: BLOCK-db! ( n ta -)
  ASM0 @ +  400 /MOD BLOCK  + C!  UPDATE ;


\ DB! is the one and only way to output code/data.
\ That way it's easy to "filter" for block output, etc.
: db! ( n ta -)
  OUTPUT @  0 = if MEM-db! else BLOCK-db! then ;
: dw! ( n ta -)
  2DUP db! CHAR+ SWAP 8 RSHIFT SWAP db! ;
: dd! ( n ta -)
  4 0 DO  2DUP db! CHAR+ SWAP 8 RSHIFT SWAP  LOOP  2DROP ;
: dc! ( n ta -)
  tW 0 DO  2DUP db! CHAR+ SWAP 8 RSHIFT SWAP  LOOP  2DROP ;

\ Data definition words (also used for code generation)
: db,  ( c -) >ASM @  SWAP OVER db!  CHAR+ >ASM ! ;
: dw,  ( x -) DUP db, 8 RSHIFT db, ;
: dd,  ( x -) 4 0 DO  DUP db, 8 RSHIFT  LOOP DROP ;
: dc,  ( x -) MODE @ if dd, else dw, then ;

\ Save memory to file
: SAVE ( "filename" -)  ( Usage: SAVE filename )
  BL PARSE ( c-addr u) R/W BIN CREATE-FILE 0<>
  if DROP ( fid) ." CREATE-FILE error" cr
  else DUP ASM0 @  $' ( location & size of code)
    ROT WRITE-FILE  0<> if ." WRITE-FILE error" cr then
    CLOSE-FILE  0<> if ." CLOSE-FILE error" cr then
  then  ;

\ For debugging - show assembly buffer:
\ TBD: only works w/ memory right now
: SHOW ( -) ASM0 @  $'  DUMP ;



variable DISP   \ displacement
variable SIB    \ -------- ssrrr00I
  \ ss=scale; rrr=index reg; I=index?
variable FLAGS  \ -------E OMIAGSDW
  \ M=r/m; I=imm; A=accumulator (AL/AX/EAX); G=seg;
  \ S=imm size (1=Short, 0=full size)
  \ W=Word or byte; E=dword (Extended reg.)
  \ O=DISP Only; D=direction (1 when r/m field is source)
variable #OPS   \ # of operands
variable #REGS  \ # of register operands
variable SRCSIZ \ size of 1st (source) operand; used for MOVZX,

: ASM-RESET  2 FLAGS ! ( D)
             0 DISP !  0 SIB !  0 #REGS !  0 #OPS ! ;
: F-SET  ( mask -) FLAGS @ OR FLAGS ! ;
: F-CLR  ( mask -) INVERT FLAGS @ AND FLAGS ! ;
: F-GET  ( mask -) FLAGS @ AND ;
: F-FLIP ( mask -) FLAGS @ XOR FLAGS ! ;

\ Set flags
: B-PTR  101 F-CLR ;
: W-PTR  100 F-CLR  1 F-SET ;
: D-PTR  101 F-SET ;

\ Read flags
: BYTE?  ( -f)  1 F-GET 1 XOR ;
: WORD?  ( -f)  1 F-GET ;   \ True for both 16-and 32-bit values
: DWORD? ( -f)  100 F-GET 0<> NEGATE ;
: CELL?  ( -f)
  100 F-GET 8 RSHIFT  MODE @  XOR INVERT  1 F-GET AND ;

: SREG?  ( -f)  8 F-GET ;

: D? ( - f)  2 F-GET ;
: S? ( - f)  4 F-GET ;
: M? ( - f) 40 F-GET ;
: O? ( - f) 80 F-GET ;

: IMM? ( -f)  20 F-GET ;
: ACC? ( -f)  10 F-GET ;
: 1REG?  #REGS @ 1 = ;
: 2REGS? ( -f) #REGS @ SREG? OR  DUP 2 =  SWAP 9 =  OR ;
               ( cc=2  OR  cc=1 + sreg )


HEX
( Conditionals )
: IF, ( opcode - a) db,  $' ( origin)  0 db, ( blank) ;
: WHILE, ( a1 opcode - a2 a1) IF, SWAP ;
: NOT, ( opcode - opcode') 1 XOR ;
: THEN, ( a -) $' OVER 1+ - SWAP db! ;
: ELSE, ( a - a') EB ( short jmp) db,
  $' OVER - SWAP db!  $'  0 db, ;
: BEGIN, ( - a) $' ;
: UNTIL, ( a opc -) db,  $'  1+ - db, ;
( REPEAT, and AGAIN, defined after JMP, )

: opc ( opcode -) ( - opcode) CREATE C, DOES> C@ ;
73 opc CS,  75 opc 0=,  79 opc 0<,  73 opc U<,  E3 opc CXNZ,
7D opc <,   7E opc >,   76 opc U>,  71 opc OV,

: LOOP,   ( a -) E2 UNTIL, ;
: LOOPZ,  ( a -) E1 UNTIL, ;
: LOOPNZ, ( a -) E0 UNTIL, ;

: FAR-JMP,  ( seg offset -)  EA db,  dc, dw,  ASM-RESET ;

HEX
: <reg> ( a - n)
  #OPS @ 1 = if $101 F-GET  SRCSIZ ! then  1 #OPS +!
  DUP 8 RSHIFT  ( high byte)
    DUP 1 AND  1 XOR  2 LSHIFT  ( S flag = NOT of bit 1)
  OR F-SET
  $FF AND ( low byte)  ;


: r/m ( n -) ( disp - n)  CREATE ,
  DOES> @ <reg> 2 F-CLR ( D) SWAP DISP !  ;
\ high byte = flags, low byte=reg
\  40000=indirect
\ 100000=disp only
OCTAL
40000 r/m [BX+SI]  40001 r/m [BX+DI]   40002 r/m [BP+SI]
40003 r/m [BP+DI]  40004 r/m [SI]      40005 r/m [DI]
40006 r/m [BP]     40007 r/m [BX]

: #@ ( disp - n)  140006 MODE @ -
                  <reg> 2 F-CLR ( D) SWAP DISP ! ;



\ Base registers: (r/m)
40000 r/m EAX@  40001 r/m ECX@  40002 r/m EDX@  40003 r/m EBX@
                                40006 r/m ESI@  40007 r/m EDI@

\ Index registers: (s-i-b, r/m=100)
: idx ( n -) ( - n)  CREATE , DOES> @
  SIB @ 1 AND if ABORT" Index reg defined twice!"
  else $FF AND SIB ! then ;
40001 idx EAX+  40011 idx ECX+  40021 idx EDX+  40031 idx EBX+
40041 idx ESP+  40051 idx EBP+  40061 idx ESI+  40071 idx EDI+

\ Scaling, for the index regs:
: 1x  300 INVERT SIB @ AND SIB ! ;
: 2x  1x  100 SIB @ OR SIB ! ;
: 4x  1x  200 SIB @ OR SIB ! ;
: 8x  1x  300 SIB @ OR SIB ! ;

\ Workaround for EBP@ and ESP@ (not really allowed as base regs)
: EBP@  #@ EBP+ ;
: ESP@  #@ ESP+ ;


: reg ( 000a00ew00rrr000 -) ( - 0000000000rrr000)  CREATE ,
  DOES> @ <reg> 1 #REGS +!  2 F-FLIP ( D)  ;
\  10000= A (accumulator)
\    400= W (word)
\ 200400= EW (dword)
210400 reg EAX  200410 reg ECX  200420 reg EDX  200430 reg EBX
200440 reg ESP  200450 reg EBP  200460 reg ESI  200470 reg EDI
 10400 reg AX      410 reg CX      420 reg DX      430 reg BX
   440 reg SP      450 reg BP      460 reg SI      470 reg DI
 10000 reg AL       10 reg CL       20 reg DL       30 reg BL
    40 reg AH       50 reg CH       60 reg DH       70 reg BH

: seg ( n -) ( -n)  CREATE , DOES> @ <reg> 2 F-SET ( D) ;
4400 seg ES   4410 seg CS   4420 seg SS   4430 seg DS
4440 seg FS   4450 seg GS

\ Debugging routine: .F (show flags, etc.)
HEX
CREATE F$  'O c, 'M c, 'I c, 'A c, 'G c, 'S c, 'D c, 'W c,
: 2^  ( n - 2^n)  1 SWAP LSHIFT ;
: .F ( -)
  BASE @ HEX
  DWORD? if 'E else '- then EMIT SPACE
  80  8 0 DO  DUP F-GET  if F$ I + C@  else '-  then
             EMIT 1 RSHIFT LOOP DROP
  #REGS @ 3 U.R  ."  regs  "
  1  SIB @ 6 RSHIFT  LSHIFT  1 U.R  ." x  DISP=" DISP @ u.
  BASE !  cr ;





: R>M ( reg - r/m)  3 RSHIFT ;

: SHORT? ( n - f)  -80 80 WITHIN ;
: # ( n1 - n1) 20 OVER SHORT? 04 AND OR F-SET ;
  \ clears all flags except I and (if short) S

: orW  ( --opc---  -  --opc--w )  1 F-GET  OR ;
: orDW ( --opc---  -  --opc-dw )  3 F-GET  OR ;




\ Generate mod r/m byte, and s-i-b byte if needed
: mod, ( md***rrr -)
  SIB @ if DUP F8 AND 4 OR db,  7 AND SIB @ OR db,
  else db,  then ;

\ Generates mod r/m byte, s-i-b byte, displacement
: modDISP, ( 00***r/m -)
  M? if
    O? if ( disp only)
      mod, DISP @ dc,
    else
      SREG? DISP @ OR  OVER 7 AND 6 = OR ( [BP])
      if
         DISP @ SWAP OVER
         SHORT? if 40 OR mod, db,  else 80 OR mod, dc,  then
      else ( zero & not seg) mod, then
    then
  else ( 11***reg) C0 OR db,  then ;

: ,IMM ( n -)
  5 F-GET 4 = if ( S,-W) db,
              else  DWORD? if dd, else dw, then  then ;









\ prefix instructions
OCTAL
: PFX CREATE C, DOES> C@ db, ;
46 PFX ES:  56 PFX CS:  66 PFX SS:  76 PFX DS:
144 PFX FS:     145 PFX GS:     ( 386+)
146 PFX OPSIZ:  147 PFX ADRSIZ: ( 386+)
362 PFX REPNZ   362 PFX REPNE
363 PFX REPZ    363 PFX REPE    363 PFX REP
\ Note: REP* was done differently in Pygmy

( Generate OPSIZ: prefix if necessary )
: (OPSIZ)  BYTE? 0=  CELL? 0=  AND  if OPSIZ: then ;




\ one-byte opcodes with no operands
: M1 ( n -) ( -)  CREATE , DOES> @ db, ASM-RESET ;
OCTAL
 47 M1 DAA,      57 M1 DAS,     67 M1 AAA,      77 M1 AAS,
324 M1 AAM,     325 M1 AAD,
220 M1 NOP,
230 M1 CBW,     230 M1 CWDE,   231 M1 CWD,
140 M1 PUSHAD,  141 M1 POPAD,  234 M1 PUSHFD,  235 M1 POPFD,
236 M1 SAHF,    237 M1 LAHF,
303 M1 RET,   \ what about RET ## form?
313 M1 RETF   \ (We shouldn't need it for Forth)
311 M1 LEAVE,
314 M1 INT3,	316 M1 INTO,
317 M1 IRET,
327 M1 XLAT,
360 M1 LOCK:
364 M1 HLT,     233 M1 WAIT,
370 M1 CLC,     371 M1 STC,    365 M1 CMC,
372 M1 CLI,     373 M1 STI,    374 M1 CLD,     375 M1 STD,

HEX
( ALU instructions with 2 operands, like ADD )
: M2 ( n -) ( various -)   CREATE ,
  DOES> (OPSIZ)  @ >R  IMM? if
    ACC? if   DROP  R> orW 4 OR db,
         else 1REG? if R>M then  80 orW db,
              R> 38 AND OR modDISP,
         then  ,IMM
  else  2REGS?  if SWAP R>M then
  	R> orDW db, OR modDISP,
  then ASM-RESET  ;
OCTAL
00 M2 ADD,  10 M2 OR,   20 M2 ADC,  30 M2 SBB,
40 M2 AND,  50 M2 SUB,  60 M2 XOR,  70 M2 CMP,


HEX
: MOV, ( r/m/i r/m - )
  (OPSIZ)
  IMM? if
    1REG? if   ( imm,reg) R>M B0 OR  WORD? 3 LSHIFT OR  db,
    else ( imm,r/m) C6 orW db, modDISP,
    then ,IMM
  else ( reg,r/m)
    O? ACC? AND
    if ( disp,acc) 2DROP A0  2 F-FLIP  orDW db,  DISP @ dc,
    else
      2REGS? if ( reg,reg) D? if SWAP then  R>M  then
      SREG? if ( sreg,r/m) 1 F-CLR 8C
      else     ( reg,r/m) 88  then
      orDW db,  OR modDISP,
    then
  then  ASM-RESET  ;

\ MOVZX, MOVSX, -- zero extend / sign extend
: MX ( n -) ( r/m reg -)
  CREATE C, C,  DOES> (OPSIZ)
  SRCSIZ @ $100 AND if ABORT" Attempt to extend a dword" then
  DUP CHAR+ C@ db,  C@ SRCSIZ @ OR db,
  2REGS? if SWAP R>M then  OR modDISP,
  ASM-RESET ;
HEX  0F B6 MX MOVZX,  0F BE MX MOVSX,


\ String instructions -- all 1-byte opcodes with W bit
: M3 ( n -) ( reg -)
  CREATE , DOES> (OPSIZ) @ orW db, ASM-RESET ;
OCTAL
246 M3 CMPS, 254 M3 LODS, 244 M3 MOVS, 256 M3 SCAS, 252 M3 STOS,


\ MUL, DIV, etc...    xxxxxxxW  mdNNNr/m
OCTAL
: M4 ( n -) ( -)  CREATE ,
  DOES> (OPSIZ)  @  366 orW db, SWAP
  1REG? if R>M then OR modDISP,  ASM-RESET ;
20 M4 COM,   40 M4 MUL,   60 M4 DIV,
30 M4 NEG,   50 M4 IMUL,  70 M4 IDIV,

\ LEA, LDS, LES instructions
OCTAL
: M5 ( n -) ( -)       \ TBD: fix if broken; add LFS/LGS/LSS
  CREATE C, DOES> (OPSIZ)  C@  db, OR modDISP, ASM-RESET  ;
215 M5 LEA,  304 M5 LES,  305 M5 LDS,

\ rotate & shift instructions
\ Note: To shift by CL, omit first operand
OCTAL
: M6 ( n -) ( n# r/m | r/m - )  CREATE ,
  DOES> (OPSIZ)  @
  320  IMM? if  3 PICK 1 <> 20 AND XOR  else  2 OR  then
  orW db,
  1REG? if SWAP R>M then OR modDISP,
  IMM? if  DUP 1 <> if db, else DROP then  then
  ASM-RESET  ;
00 M6 ROL,  20 M6 RCL,	40 M6 SHL,
10 M6 ROR,  30 M6 RCR,  50 M6 SHR,  70 M6 SAR,








\ INC, DEC instructions
OCTAL
: M7 ( opc -) ( reg | r/m - )  CREATE ,
  DOES> (OPSIZ)  @ SWAP  1REG? if ( opc reg) R>M then
  1REG?  WORD? AND  ( full-size register?)
  if ( opc rX)  OR 100 OR db,
  else ( opc mem | opc rH | opc rL )
    376 orW db,  OR modDISP,
  then  ASM-RESET  ;
00 M7 INC,   10 M7 DEC,






\ PUSH, POP instructions
HEX
: M8 ( n -) ( reg | seg | r/m -)  CREATE ,
  DOES> @  8 ( G) F-GET
  if ( seg opc) OVER 20 AND
    if   ( sreg3) 0F db,  4 RSHIFT 1 AND 1 XOR  80 OR  OR db,
    else ( sreg2)         4 RSHIFT 1 AND 1 XOR   6 OR  OR db,
    then
  else 1REG?
    if ( reg opc) 2/  8 AND  8 XOR  50 OR  SWAP R>M OR db,
    else ( r/m opc) DUP 8 RSHIFT  FF AND  db,  OR modDISP,
    then  then  ASM-RESET  ;
FF30 M8 PUSH,    8F00 M8 POP,   \ Hex
\ 177460 M8 PUSH,  107400 M8 POP, \ Octal


\ IN, OUT instructions
OCTAL
: M9 ( n -) ( n# r1 | r1 -)  CREATE ,
  DOES> @  (OPSIZ) orW NIP
    IMM? if ( n# opc)  db, ( n#) else ( opc) 10 OR then db,
    ASM-RESET  ;
344 M9 IN,   346 M9 OUT,

\ XCHG
HEX
: XCHG,  ( reg mem | mem reg | reg1 reg2 -)
  (OPSIZ)
  #REGS @ 2 =  ACC? AND  WORD? AND ( AX and another reg)
  if  ?DUP if NIP then ( r1) R>M  90 OR db,
  else 2REGS? if  R>M  then  OR  86 orW db, modDISP,
  then  ASM-RESET  ;

\ TEST -- sim. to ADD
HEX
: TEST,  ( various -)
  (OPSIZ)
  IMM?
  if ACC? if     DROP  A8 orW ( 4 OR) db,
          else   1REG? if R>M then  F6 orW db,
                 ( OR)  modDISP,
          then  ,IMM
  else  2REGS?  if SWAP R>M then
        84 orW db, OR modDISP,
  then  ASM-RESET  ;



\ INT -- usage: 21 INT,
OCTAL
: INT, ( #n -) 315 db, db, ASM-RESET ;

\ CALL
HEX
: CALL, ( various -)  IMM?	( intra-seg direct)
  if  ( n#)  $ 1+ tW + -  ( relative)
    E8 db, dc,   ( usage: 2338 # CALL, )
  else   ( intra-seg indirect: address in reg/mem)
    ( mem | reg -)  1REG?  if R>M then
    FF db,  10 OR modDISP, ( usage: 0 [BX] CALL,  or  DX CALL, )
  then  ASM-RESET  ;





\ JMP
HEX
: JMP, ( various -) 140 F-GET  ( R or M intra-seg indirect)
  if ( mem | reg -)  1REG?  if R>M then
    FF db,  20 OR modDISP,  \ 0 [BX] JMP,   DX JMP,
                            \ 3759 #@ JMP,
  else
    ( a) $ 1+ -  ( relative)
    DUP SHORT? if   1-      EB db, db,
               else tW - E9 db, dc, then
  then  ASM-RESET  ;

HEX
: LJMP, ( a -)  E9 db,  $  tW + - dc, ;

: AGAIN, ( a -)  -ORG @ + JMP, ;
: REPEAT, ( a a -) AGAIN, THEN, ;


\ Some 386/486 instructions...
OCTAL
: CLTS,		17 db, 06 db, ;
: INVD,		17 db, 10 db, ;
: WBINVD,	17 db, 11 db, ;

OCTAL
: LABEL: ( "name" -) ( - a) CREATE $ , DOES> @ ;
: LABEL' ( "name" -) ( - a) CREATE $' , DOES> @ ;

: JMP-FROM ( "name" -) 351 db,  LABEL'  0 dc, ;
: JMP-TO ( a -)
  $'  OVER -  tW - ( displacement)
  SWAP dc! ;


: 0FILL ( # -  Pad out target file w/ 0's)
  ?DUP if  0 DO 0 db, LOOP  then ;

\ Pad w/ NOPs
: ALIGN, ( u -)  $  OVER MOD
  DUP if - 0 DO  NOP,  LOOP
  else 2DROP then ;

\ Pad w/ zeroes
: 0ALIGN, ( u -)  $  OVER MOD
  DUP if - 0 DO  0 db,  LOOP
  else 2DROP then ;

\ DPUSH/DPOP pseudo-instructions
\ Use this like: 32 # DPUSH,  or  EDX DPOP,
: 0PUSH, ( -  Prepare a DPUSH, [equivalent to DUP])
  $83 db, $ED db, tW db,  ( tW # EBP SUB,)
  $89 db, $45 db, $00 db, ;  ( EAX 0 EBP@ MOV,)
: DPUSH, ( r/m - )
  0PUSH,
  ( r/m) EAX MOV,  ;
: DPOP, ( r/m - )
  EAX  SWAP ( r/m) MOV,
  0 EBP@ EAX MOV,
  tW # EBP ADD,  ;
: DROP, ( - )
  0 EBP@ EAX MOV,
  tW # EBP ADD, ;

\ Debugging routine:
: $.  BASE @ HEX $ . DECIMAL $' . BASE ! CR ;

\ Locks the CPU (purposely of course)
: HANG  $EB db, -2 db, ;


ASM-RESET
HEX USE16
