\ forth.f - high(er) level Forth words          29 Mar 2000  tcn

DECIMAL
10 tval base
: hex     16 BASE ! ;
: decimal 10 BASE ! ;
: octal    8 BASE ! ;
: binary   2 BASE ! ;

$2000 tval tib   \ address of Terminal Input Buffer
\ tvar #tib  \ length of TIB (current, not max.)
\ tvar >in   \ current position in input buffer

\ Dictionary sections (pointers to the linked lists):
\ tCREATE dict  0 dc, ( Main)  0 dc, ( Compiler)
tVAR dict


\ "Classic version" - should work w/ signed & unsigned #'s
: WITHIN ( n min max - ?) OVER - PUSH - POP U< ;


\ Screen output
\ -------------

32 tCON bl
: space   32 EMIT ;
: cr      10 EMIT ;
: spaces ( # - )  0 MAX  ?DUP IF  FOR SPACE NEXT  THEN ;

\ Number output (from Pygmy Forth)
: HOLD ( ..# x n - ..# x)  SWAP PUSH SWAP 1+  POP ;
: DIGIT ( n - c)  DUP 9 >  7 AND +  48 + ;
: <# ( n - ..# n)  0 SWAP ;
: #> ( ..# n -)   DROP FOR EMIT NEXT ;
: SIGN  (  ..# n n - ..# n)  0< IF  45 HOLD   THEN ;
: # ( ..# n - ..# N)  BASE @ U/MOD  SWAP DIGIT HOLD ;
: #S  ( ..# n - ..# 0)  BEGIN  #  DUP 0= UNTIL ;
: (.)  ( n - ..# n)   DUP PUSH ABS  <# #S  POP SIGN ;
: .r ( n n -)  PUSH  (.) OVER POP SWAP -  SPACES #> ;
: . ( n -)    0 .R  SPACE ;
: u.r ( u n -)  PUSH  <# #S  OVER POP SWAP -  SPACES #> ;
: u. ( u -)   0 U.R  SPACE  ;

: ? ( a)  @ . ;
: base.  BASE @ DECIMAL DUP . BASE ! ;  \ from R. Hohense

\ : .S ( - )  '< EMIT DEPTH DUP . '> EMIT SPACE
\ TBD: write DEPTH, finish this.
\   ;

: type ( a # - )
  ?DUP IF  FOR C@+ EMIT NEXT DROP  THEN ;

\ String literals (Metacompiler version):
{ :' sliteral  ( a #) ( - cstr)  }'
  $E8 db, DUP 1+ dc,                     \ JSR to end of string
  DUP db, 0 DO dup c@ db, char+ LOOP drop \ Store string
  0push, eax pop, ;
{ :' "  ( - cstr )  '" PARSE ( a #) sliteral }' ;
{ :' ." ( -)        " COUNT TYPE }' ;

: at-xy  ( x y - (Position the cursor)
  80 * + 2* [[ ax XY #@ mov, >XY # call, drop, ]] ;

\ Keyboard input
\ --------------

: key? ( -f)  KBUF DUP @  SWAP w+ @  <>  ;
: key ( -c)
  BEGIN  PAUSE  KEY?  UNTIL   \ Idle loop.. wait for key
  KBUF w+ @  DUP C@  SWAP 1+  \ Get key
  KBUF 12 + @  OVER = IF  DROP  KBUF 8 + @  THEN  \ wrap
  KBUF w+ ! ;  \ Update tail


\ : ACCEPT ( a # - #'  (Read up to # chars, actually read #')
\  SWAP OVER PUSH
\  BEGIN OVER WHILE   KEY DUP 10 - WHILE
\   DUP 8 = IF DROP   OVER R@ <
\              IF 1- 32  OVER C!  1 +UNDER
\                    8 EMIT  SPACE  8 EMIT
\              THEN
\           ELSE   DUP EMIT OVER C!  1 neg +UNDER 1+
\           THEN
\  REPEAT DROP SPACE THEN  DROP  POP SWAP -
\  ;


\ String handling
\ ---------------

tCODE: /string ( a # u - a # )
  eax 0 ebp@ sub, eax 4 ebp@ add, drop, ;tCODE


\ Interpreter subroutines
\ -----------------------

\ Number parsing:
tCREATE bases   16 db, 2 db, 10 db, 255 db,

: -digit ( c - n 1 | 0 )
  base @ 255 = IF true exit THEN
  dup 57 u> IF 95 and 7 - THEN 48 -
  dup base @ u>= IF drop false exit THEN true ;

: 10*+ ( n -  (Multiplies by BASE and adds digit)
  here @  base @ * +  here ! ;

: >num ( a # - n 1 | 0 )
  true here w+ !
  0 here !
  FOR
    dup c@ -digit IF 10*+ ELSE false here w+ ! THEN
    1+ NEXT drop
  here w+ @ dup IF here @ swap THEN ;

 : s>num  ( a # - n 1 | 0 )
  base @ push  \ Save BASE
  over c@ 45 = dup push IF 1 /string THEN  \ Negative?
  over c@ 36 - dup 4 u<   \ Get BASE prefix ($ % & ')
    IF bases + c@ base ! 1 /string
    ELSE drop THEN
  >NUM
  dup pop and IF swap neg swap THEN
  pop base ! ;


: words
  DICT @  BEGIN
    DUP  1-  DUP C@  SWAP OVER - SWAP
    TYPE SPACE
    @ ?DUP 0=
  UNTIL CR ;

tCODE: find ( a # - xt -1 | 0 )
  ebx push, ecx push, edx push, esi push, edi push,
  { DICT } 5 + # edx mov,   \ edx = header
  ebx ebx sub,          \ ebx = 0 if no match
  BEGIN,
    ecx ecx sub, ecx dec,
    ebx ebx test, 0=,
      IF,   0 edx@ edx mov,  \ Go to next word
      ELSE, ecx inc,
    THEN, edx ecx and, ecx ecx test, 0=, NOT,
  WHILE,
    -1 edx@ al cmp,  0=, IF,  \ compare lengths
      0 ebp@ esi mov,  \ search string
      edx edi mov, edi dec, eax edi sub,     \ string in dict
      eax ecx mov, cld, rep cmps, 0=, IF, ebx dec, THEN,
    THEN,
  REPEAT,
  ebx ebx test, 0=, IF, drop,
                  ELSE, 4 # edx add, edx 0 ebp@ mov, THEN,
  ebx eax mov,
  edi pop, esi pop, edx pop, ecx pop, ebx pop,  ;tCODE

: sp.  [[ 0push,  ebp eax mov, ]] . ;
: 5.   . . . . . ;

: CForth ( -  (Beginnings of a Color Forth interpreter)

  BEGIN
    TIB @
    BEGIN  KEY DUP EMIT
           DUP 32 <>  OVER 10 <>  AND
    WHILE  OVER C! 1+  REPEAT
    DROP  TIB @  SWAP OVER -
    ?DUP IF  \ Not zero-length
      2DUP FIND
      IF    NIP NIP  EXECUTE
      ELSE  S>NUM  NOT IF '? EMIT CR THEN
      THEN
    ELSE DROP THEN
  AGAIN ;



\ : COMPARE ( a # a # - n  (0 if equal, otherwise +1/-1)
\ Case-insensitive compare:
\   Convert chars w/ XLAT.  Lower->upper, whitespace->32.
\   ;

\ : EVALUATE ( a # - )
\ save input specs
\  0 >IN !
\  BEGIN
\    BL WORD FIND ?DUP
\    IF   
\    ELSE 0 0 ROT COUNT >NUMBER
\      IF   ( push/literal)
\      ELSE ( error)
\      THEN
\    THEN
\  ( end of string?) UNTIL
\ restore input specs
\  ;


\ : CREATE ( "name" - ) ( - PFA )
\   HEAD HERE ,
\  ;

\ : DOES>
\
\  ;
