\ loader.f - Boot sector for non-fs disk       7 Oct 1999  tcn
\
\ Mission:
\   Load kernel, starting from the 2nd sector of the boot drive,
\      at address 0x10000.
\   Enable A20 and enter protected mode (if room)
\
\ Limitations:
\   Kernel must be 64kb maximum (this allows a simplification)
\   (Not a problem for a FORTH kernel)






USE16
JMP-FROM start:

DECIMAL
  \ Size of kernel, in sectors (default to max=64k=128)
  LABEL' #sec'  LABEL: #sec    128 db,










LABEL: read-sect    HEX
\ Input: DL=drive, CL=sector, AL=count, ES=destination (ES:0)
\ * Limitations: 128 sectors max; must start on track 0
\ * Uses multitrack reads (some crappy BIOSes can't handle that)
\   ...seems to work with up to 72 sectors in kernel (36 blocks)
  DX DX XOR, ( DL=drive=0, DH=head=0)   CH CH XOR, ( CH=track=0)
  BX BX XOR, ( ES:BX=address)
  2 # AH MOV, ( AH=command)  13 INT,
  CS, NOT, ( not carry) IF, RET, THEN,
\ Arrive here on error...
\  AH AH XOR,  16 INT, ( wait for keypress)
  E00 'E + # AX MOV,  7 # BX MOV,  10 INT,
\  19 INT,             ( reboot)
\  E00 '* + # AX MOV,  7 # BX MOV,  10 INT,
  hang

\ TBD: <support routines here>















start: JMP-TO   HEX
  AX AX XOR,    AX DS MOV,  AX SS MOV,
  7C00 # SP MOV,  SP BP MOV,
  \ Note: DL=boot drive

\ '1 $E00 + # AX MOV,  7 # BX MOV,  10 INT,
  \ Load boot sector extension (if needed) from 2nd sector
  7E0 # AX MOV,  AX ES MOV,  2 # CL MOV,  1 # AL MOV,
  read-sect # CALL,

\ '2 $E00 + # AX MOV,  7 # BX MOV,  10 INT,
  \ Load kernel: Up to 64 blocks (128 sec) from 3rd sector
  1000 # AX MOV,  AX ES MOV,  3 # CL MOV,
  #sec #@ AL MOV,
  read-sect # CALL,

  \ Turn off floppy motor
  3F2 # dx mov,  al al xor,  al out,

\ TBD: <Squeeze in some initialization stuff here>
\ checking, pmode, paging..

\ Enter the kernel..
  EA db,  0 dw,  1000 dw, ( JMP FAR 1000:0000)

\ End of boot sector - check size & add signature
DECIMAL
$'
\ 50
." Boot sector is " DUP . ." bytes ("
510 SWAP -  DUP . ." free)" CR
DUP 0< [IF]  ." Trim it down!" CR BYE [THEN]
0FILL ( to byte 510)
$55 db, $AA db, ( Signature)

\ --------------------------------------------------------------
\ Sector 2 -- Available for boot sector extension
\   Use this space to make it more robust...
\    - workarounds for crappy BIOSes
\    - more useful error messages
\ --------------------------------------------------------------

\ ...
\ <code here>
\ ...






\ End of 2nd sector
DECIMAL
$' 512 -
." 2nd sector is " DUP . ." bytes ("
512 SWAP - DUP . ." free)" CR
DUP 0< [IF]  ." Trim it down!" CR BYE [THEN]
0FILL









